#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <getopt.h>
#include <signal.h>
#include <sys/stat.h>
#include <scsi/sg.h>
#include "scsi_cmds.h"
#include "wave.h"

extern int optind;
extern char *optarg;

int g_lead=60,g_gap=0,g_write,g_rate=48000,done=0,g_endid=0,g_help=0;
char g_outfile[100],g_dev[100]="\0";
int write_st,track_no,sample_rate,new_prog;
char hr,min,sec,fr;
char p_hr,p_min,p_sec,p_fr;

int bcd2l(int x){
  return (((x)&0xf)+(((x)&0xf0)>>4)*10+(((x)&0xf00)>>8)*100+(((x)&0xf000)>>12)*1000);
}

int l2bcd(int x){
  return (((x)%10)+(((x)/10)%10)*0x10+(((x)/100)%10)*0x100+(((x)/1000)%10)*0x1000);
}

int bcdinc(int x){
  return l2bcd(bcd2l(x)+1);
}

void int_func(int dud)  //Interrupt handler for CTRL-C signal
{
  done=1;
  return;
}

int parse_options(int argc,char *argv[])
{
  int i;

  while((i=getopt(argc,argv,"l:g:r:o:d:ewh"))!=EOF){
    switch(i){
      case 'h':
        g_help=1;
        break;
      case 'd':
        strncpy(g_dev,optarg,99);
        g_dev[99]=0;
        break;
      case 'e':
        g_endid=1;
        break;
      case 'r':
        g_rate=atoi(optarg);
        break;
      case 'l':
        g_lead=atoi(optarg);
        break;
      case 'g':
        g_gap=atoi(optarg);
        break;
      case 'o':
        strncpy(g_outfile,optarg,99);
        g_outfile[99]=0;
        break;
      case 'w':
        g_write=1;
        break;
      default:
        g_help=1;
    }
  }
  return optind;
}

void usage(char *progname)
{
  fprintf(stderr,
"\nUsage: %s [options] [[input file] [input file] ...] }\n"
"Options:\n"
"\t-h\tThis help menu.\n"
"\t-d dev\tSet the SCSI generic device (default is /dev/sga).\n"
"\t-w\tWrite Mode (default is \"read\" from tape).\n"
"\t-o fn\tSpecify the output file, fn, in read mode (default is STDOUT).\n"
"\t-l n\tSpecify the length, n, in seconds for a \"lead-in\".\n"
"\t-r n\tSpecify the sample rate, n, for writing (44100 or 48000)\n"
"\t-g n\tSpecify a length, n, in seconds for track gaps.\n"
"\t-e\tWrite an END-ID at end of tape.\n"
"\n"
"Obviously, different options apply for read and write operations.\n"
"RIFF (wave) files, or RAW little-endian files are expected for input.\n"
"Output file is written as RIFF (wave).\n"
  ,progname);
  exit(0);
}

int set_density(int fd,int d_code,int blksz)
{
  char buf[12];
  int i;

  reset_device(fd);
  memset(buf,0,12);
  buf[2]=0x10;
  buf[3]=8;
  buf[4]=d_code;
  buf[9]=(blksz>>16)&0xff;
  buf[10]=(blksz>>8)&0xff;
  buf[11]=blksz&0xff;
  mode_select(fd,buf,12);
  memset(buf,0,12);
  buf[3]=8;
  mode_sense(fd,buf,12);
  i=buf[4]&0xff;
  if(i!=0x80){
    return 0;
  }
  return 1;
}

//44K 00:00:54:21 AA040000 0400  Sharp codes
//44K 00:00:54:22 0304C000 0400


unsigned char *build_sub(unsigned char *block)
{
  unsigned char *p;

  p=block;
  //Recorded date/time--fixme  Hard Coded (does any deck utilize this anyway?)
  p[5760+32]=0x57;  //  Hi: 5 == pac type "Date" ; Low 7 == day of week "Sunday"
  p[5760+33]=0x01;  //Year 
  p[5760+34]=0x06;  //Mon
  p[5760+35]=0x21;  //Day
  p[5760+36]=0x19;  //Hour
  p[5760+37]=0x28;  //Min
  p[5760+38]=0x13;  //Sec
  p[5760+39]=0x53;  //Parity 

  p[5760]=0x10;       //Prog time
  p[5760+1]=0x02;
  p[5760+2]=0;

  p[5760+8]=0x20;     //ABS time
  p[5760+9]=0x02;
  p[5760+10]=0;

  p[5760+16]=0x30;      //Record time
  p[5760+17]=0x02;
  p[5760+18]=0;

  fr++;                          //0 sec frames 0-32
  if((fr>(32+((sec%3)==2)))){     //1 sec frames 0-32
    fr=0;                        //2 sec frames 0-33
    sec++;
    if(sec==60){
      sec=0;
      min++;
      if(min==60){
        min=0;
        hr++;
      }
    }
  }
  p_fr++;                          //0 sec frames 0-32
  if((p_fr>(32+((p_sec%3)==2)))){     //1 sec frames 0-32
    p_fr=0;                        //2 sec frames 0-33
    p_sec++;
    if(p_sec==60){
      p_sec=0;
      p_min++;
      if(p_min==60){
        p_min=0;
        p_hr++;
      }
    }
  }
  p[5760+3]=l2bcd(p_hr);  
  p[5760+4]=l2bcd(p_min);  
  p[5760+5]=l2bcd(p_sec);  
  p[5760+6]=l2bcd(p_fr);  
  p[5760+7]=p[5760]^p[5760+1]^p[5760+2]^p[5760+3]^p[5760+4]^p[5760+5]^p[5760+6];

  p[5760+11]=p[5760+19]=l2bcd(hr);
  p[5760+12]=p[5760+20]=l2bcd(min);
  p[5760+13]=p[5760+21]=l2bcd(sec);
  p[5760+14]=p[5760+22]=l2bcd(fr);

  p[5760+15]=p[5760+8]^p[5760+9]^p[5760+10]^p[5760+11]^p[5760+12]^p[5760+13]^p[5760+14];
  p[5760+23]=p[5760+16]^p[5760+17]^p[5760+18]^p[5760+19]^p[5760+20]^p[5760+21]^p[5760+22];
  if(write_st){
    if(!new_prog){
      new_prog=1;
      p_hr=p_min=p_sec=p_fr=0;
    }
    *((unsigned long *)(p+5815))=0x0004C000|(l2bcd(track_no)<<24);
    write_st--;
    if(write_st==0){
      track_no++;
      new_prog=0;
    }
  }else{
    *((unsigned long *)(p+5815))=0xAA040000;
  }
  *((unsigned short *)(p+5819))=(sample_rate==44100?1:0)<<10;
  return p;
}

//BB028000 Sharp lead-in sub-id

int write_lead(int fd,unsigned char *block)
{
  unsigned char *p;

  hr=min=fr=0;
  sec=3;
  p=block;
/*                          Sony SDT-9000's freak if this code is used
  p[5760+8]=0x20;
  p[5760+9]=0x01;
  p[5760+10]=0x00;
*/
  *((unsigned long *)(p+5760+56))=0x00BB0080;
  *((unsigned short *)(p+5820))=0x0004;//(sample_rate==44100?1:0)<<10;
  //for(sec=3;sec;sec--){
    for(fr=100;fr;fr--){
/*                          Sony SDT-9000's freak if this code is used
      p[5760+11]=l2bcd(hr);
      p[5760+12]=l2bcd(min);
      p[5760+13]=l2bcd(sec);
      p[5760+14]=l2bcd(fr);
      p[5760+15]=p[5760+8]^p[5760+9]^p[5760+10]^p[5760+11]^p[5760+12]^p[5760+13]^p[5760+14];
*/
      write_6(fd,p);
      fprintf(stderr,"\rFormat:  %dK %.2x:%.2x:%.2x %.8lX %.4X",(int)sample_rate/1000,p[5760+11],p[5760+12],p[5760+13],*((unsigned long *)(p+5816)),*((unsigned short *)(p+5820)));
      //write_6d(fd,p);
    }
  //}
  fprintf(stderr,"\n");
  return 1;
}

int check_file_magic(FILE *f)
//return 0 if raw, or RATE if wav
{
  riff head;
  
  fread(&head,1,sizeof(riff),f);
  if(head.magic==0x46464952){ //RIFF
    //fseek(f,,SEEK_SET);
    return head.samp_per_sec;
  }
  fseek(f,0L,SEEK_SET);
  return 0;

}

int detect_drive(void)
//fixme  Need detection method.  Hard coded to /dev/sga
{
  int fd;

  if(g_dev[0]){
    fd=open(g_dev,O_RDWR);
  }else{
    fd=open("/dev/sga",O_RDWR);
  }
  if(fd<0){
    fprintf(stderr,"Need read/write permissions for %s.\n",g_dev[0]?g_dev:"/dev/sga");
    exit(1);
  }
  return fd;
}

int main(int argc,char *argv[])
{
  unsigned char p[5822],tmpbuf[5822];
  unsigned long *subid,outdatasize,wave_rate,fssec;
  unsigned short *mainid;
  int fd,count,write=0,fcount,x,opi,ff,err_code,chunksz,partial_read=0,readsz,a_err=0,b_err=0;
  size_t i;
  FILE *f;
  struct stat fsbuf;
  //44.1=5292 48=5760  sr=1 for 48K or 0 for 44K

  opi=parse_options(argc,argv);
  if(g_help){
    usage(argv[0]);
  }
  fcount=argc-opi;
  write=g_write;
  if(write){
    fprintf(stderr,"Writing %d files with %d sec. lead-in and %d second gaps.\n",fcount,g_lead,g_gap); 
  }
  //return 1;
  //if(argc>1)write=1;
  fd=detect_drive();
  if(!set_density(fd,0x80,0)){
    fprintf(stderr,"Unable to set device to AUDIO mode.\n");
    exit(1);
  }
  new_prog=write_st=0;
  track_no=1;
  sample_rate=g_rate;
  if(write){
    memset(p,0,5822);
    subid=(unsigned long *)(p+5760+56);
    mainid=(unsigned short *)(p+5760+60);
    //write_st=333;
    write_lead(fd,p);
    hr=min=sec=fr=0;
    memset(p,0,5822);
    count=(int)(33.3*g_lead);
    for(x=0;x<count;x++){    //write lead-in
      build_sub(p);
      write_6(fd,p);
      //write_6d(fd,p);
      fprintf(stderr,"\rLead-in: %dK %.2x:%.2x:%.2x %.8lX %.4X",sample_rate/1000,p[5760+11],p[5760+12],p[5760+13],*subid,*mainid);
    }
    for(ff=opi;ff<argc;ff++){
      f=fopen(argv[ff],"r");
      if(f==NULL){
        perror("");
        exit(1);
      }
      wave_rate=check_file_magic(f);
      fstat(fileno(f),&fsbuf);
      if(wave_rate){
        fssec=(fsbuf.st_size-44)/(wave_rate*4); //fssec/3600,(fssec%3600)/60,fssec%60
        fprintf(stderr,"\n%s is a RIFF (wave) file.  Rate=%d %.2d:%.2d:%.2d\n",argv[ff],(int)wave_rate,(int)fssec/3600,(int)(fssec%3600)/60,(int)fssec%60);
        sample_rate=wave_rate;
      }else{
        fssec=(fsbuf.st_size)/(sample_rate*4);
        fprintf(stderr,"\nAssuming little endian RAW audio file at rate %d %.2d:%.2d:%.2d\n",(int)sample_rate,(int)fssec/3600,(int)(fssec%3600)/60,(int)fssec%60);
      }
      chunksz=(sample_rate==48000)?5760:5292;
      write_st=333;
      memset(p,0,5760);     //clear the buffer
      while(!feof(f)){
        if(partial_read){
          memcpy(p,tmpbuf,partial_read);
          readsz=fread(p+partial_read,1,chunksz-partial_read,f)+partial_read;
          fprintf(stderr,"\ntacked on %d bytes to make %d total\n",partial_read,readsz);
          partial_read=0;  //reset
        }else{
          readsz=fread(p,1,chunksz,f);
        }
        if(readsz<chunksz){
          partial_read=readsz;
          fprintf(stderr,"\npartial read: %d bytes\n",partial_read);
          memcpy(tmpbuf,p,5822);
          break;            //we'll tack it onto the next block
        }
        build_sub(p);
        write_6(fd,p);
        //write_6d(fd,p);
        fprintf(stderr,"\rWriting: %dK %.2d:%.2d:%.2d %.8lX %.4X",sample_rate/1000,hr,min,sec,*subid,*mainid);
      }
      if(g_gap){
        memset(p,0,5760);
        if(partial_read){                //flush remaining bytes
          memcpy(p,tmpbuf,partial_read);
          build_sub(p);
          write_6(fd,p);
          partial_read=0;  //reset
        }
        memset(p,0,5760);    //Whoops.  Wasn't resetting the block if there was flush data.
        count=(int)(33.3*g_gap);
        for(x=0;x<count;x++){    //write gap
          build_sub(p);
          write_6(fd,p);
          //write_6d(fd,p);
          fprintf(stderr,"\rGap:     %dK %.2x:%.2x:%.2x %.8lX %.4X",sample_rate/1000,p[5760+11],p[5760+12],p[5760+13],*subid,*mainid);
        }
      }
      fclose(f);
    }
    if(partial_read){                //flush any remaining bytes
      fprintf(stderr,"\ntacking on %d bytes from partial read\n",partial_read);
      memset(p,0,5760);
      memcpy(p,tmpbuf,partial_read);
      build_sub(p);
      write_6(fd,p);
      partial_read=0;  //reset
    }
    //            fixme:  write a end-id (is this a proper END-ID subcode?)
    if(g_endid){
      write_st=count=333;
      for(x=0;x<count;x++){    //write leadout (end-id)
        build_sub(p);
        *((unsigned long *)(p+5815))=0x0004C000|(0xEE)<<24;
        write_6(fd,p);
        //write_6d(fd,p);
        fprintf(stderr,"\rLead-out:%dK %.2x:%.2x:%.2x %.8lX %.4X",sample_rate/1000,p[5760+11],p[5760+12],p[5760+13],*subid,*mainid);
      }
    }
    fprintf(stderr,"\n");
  }else{
    if(g_outfile[0]){
      f=fopen(g_outfile,"w+");
      if(f==NULL){
        perror("File I/O error:");
        return 1;
      }
    }else{
      f=stdout;
    }
    write_wave_header(f,sample_rate);
    outdatasize=0;
    signal(SIGINT,int_func);
    fprintf(stderr,"\nReading from tape.  CTRL-C to exit and save file.\n");
    while(!done){
      read_6(fd,p);
      if(p[5820]&0x04){
        if(sample_rate!=44100){
          fprintf(stderr,"\n Sample rate changed from %dK to 44.1K (File position: %lu)\n",sample_rate/1000,outdatasize);
          sample_rate=44100;
        }
      }else{
        if(sample_rate!=48000){
          fprintf(stderr,"\n Sample rate changed from %dK to 48K (File position: %lu)\n",sample_rate/1000,outdatasize);
          sample_rate=48000;
        }
      }
      i=fwrite(p,1,(sample_rate==48000)?5760:5292,f);
      outdatasize+=i;
      if(outdatasize>=0xFFFFFFFE){
        done=1;
      }
      //fwrite(p+5760,62,1,stdout);
      //fprintf(stdout,"%c%c",0,0);   
      subid=(unsigned long *)(p+5760+56);
      mainid=(unsigned short *)(p+5760+60);
      err_code=p[5760+59];
      hr=p[5760+11];
      min=p[5760+12];
      sec=p[5760+13];
      fr=p[5760+14];
      //fprintf(stdout,"%dK %.2d:%.2d:%.2d:%.2d %.8lX %.4X\n",sample_rate/1000,bcd2l(hr),bcd2l(min),bcd2l(sec),bcd2l(fr),*subid,*mainid);
      fprintf(stderr,"\r%dK ATIME: %.2d:%.2d:%.2d SUBID: %.8lX MAINID: %.4X ERR A:%.2d B:%.2d",sample_rate/1000,bcd2l(hr),bcd2l(min),bcd2l(sec),*subid,*mainid,a_err,b_err);
      if(err_code&0x20){
        a_err++;
        fprintf(stderr,"\n Read error detected on head A at (%.2d:%.2d:%.2d Fr:%2d) file position: %lu\n",bcd2l(hr),bcd2l(min),bcd2l(sec),bcd2l(fr),outdatasize);
        //break;
      }
      if(err_code&0x40){
        b_err++;
        fprintf(stderr,"\n Read error detected on head B at (%.2d:%.2d:%.2d Fr:%2d) file position: %lu\n",bcd2l(hr),bcd2l(min),bcd2l(sec),bcd2l(fr),outdatasize);
        //break;
      }
    }
    fprintf(stderr,"\nDone.  Writing RIFF/WAV header\n");
    fixup_wave_header(f,outdatasize,sample_rate);         //sample rate may have changed.
    fclose(f);
    return 0;
  }  
  return 0;
}

